/*
 * (C) Copyright Pattern Recognition Group Delft. All rights reserved
 *
 * QUADR.C : Calculate a quadratic discrimination function with the Fisher
 *            criterium.
 *
 * References: Duin R.P.W, "Scheiding van Puntverzamelingen", Afstudeer verslag,
 * Technische Hogeschool Delft, Juni 1970. Appendix A.
 *
 * Made by :  W.F. Schmidt
 * SccsId  :  %Z%$RCSfile: quadr.c,v $ V%Q% $Revision: 1.4 $ $Date: 1996/10/07 06:50:24 $
 *
 * Modified by	: H.J. Huisman & A. Hoekstra
 * Purpose	: Add a non-linear part to fisher
 */

#define QUADR
#define FQTINY 1e-10

#include <stdio.h>
#include <math.h>

#include <sprlib.h>

/*
 * WEIGHT0_VECTOR : Calculate the constant value of the descision surface for
 * the quadratic Fisher discriminant function.
 * Kainv, Kbinv, Ma, Mb, c and n, deta and detb are input. b is the output.
 */

static double weight0_vector (double **Kainv, double **Kbinv, double *Ma, double *Mb, double c, int n, double deta, double detb)
{
	int i;
	double *bha, *bhb, b;

	if ((bha = vector(1, n)) == NULL)
		return TRUE;
	if ((bhb = vector(1, n)) == NULL)
		return TRUE;

	matvecmult(Kainv, n, n, Ma, bha);
	matvecmult(Kbinv, n, n, Mb, bhb);

	b = 0.0;
	for (i = 1; i <= n; i++)
		b += bhb[i] * Mb[i];
	for (i = 1; i <= n; i++)
		b -= bha[i] * Ma[i];

	free_vector(bhb, 1, n);
	free_vector(bha, 1, n);

	if (deta < FQTINY)
	{
		sprerror("fisherq(): Division by zero attempted.");
		return -1.0;
	}

	b += 2 * log(c) + 2 * log(deta / detb);

	return b;
}



/*
 * FISHERQ : Calculate the optimal quadratic descision surface for multi-normal
 * distributed classes. This is an implementation of the Fisher linear
 * discrimination function, which has been extended with a quadratic part.
 * On input Ka[1..n][1..n] and Kb[1..n][1..n] are
 * the covariances of the two classes. Ma[1..n] and Mb[1..n] are respectively
 * both mean values of the different classes. N is the dimension of the
 * feature space. C is an extra parameter and is defined as:
 *
 * C = P(b)/P(a);
 *
 * Where P(b) and P(a) are the apriori changes to select a member of class b and
 * class a respectively. The descision function is modeled as  W*x + b = 0,
 * where b is the constant term. On output  W1[1..n] and b are set to the
 * optimal value.
 *
 * REF : See Duin appendix A.
 *
 * Return value : TRUE  If a math error is detected
 *                FALSE If a valid fisher vector is found.
 */

int fisherq (double **Ka, double **Kb, double *Ma, double *Mb, double c, int n, double **W2, double *W1, double *b)
{
	double **Kainv, **Kbinv, deta, detb;
	int i, j;

	if ((Kainv = matrix(1, n, 1, n)) == NULL)
		return TRUE;
	if ((Kbinv = matrix(1, n, 1, n)) == NULL)
		return TRUE;

	matinvd(Ka, n, Kainv, &deta);
	matinvd(Kb, n, Kbinv, &detb);

	/*
	 * Calculate the quadratic matrix W2 and calculate the normal vector
	 * W1.
	 */
	for (i = 1; i <= n; i++)
	{
		W1[i] = 0.0;
		for (j = 1; j <= n; j++)
		{
			W2[i][j] = Kbinv[i][j] - Kainv[i][j];
			W1[i] += 2 * (Kainv[i][j] * Ma[j] - Kbinv[i][j] * Mb[j]);
		}
	}

	/* Calculate the constant term b. */
	if ((*b = weight0_vector(Kainv, Kbinv, Ma, Mb, c, n, deta, detb)) == -1.0)
	{
		free_matrix(Kainv, 1, n, 1, n);
		free_matrix(Kbinv, 1, n, 1, n);

		return TRUE;
	}

	free_matrix(Kainv, 1, n, 1, n);
	free_matrix(Kbinv, 1, n, 1, n);

	return FALSE;
}

/*
 * FISHERQ_DATASET : Calculate the optimal quadratic descision surface for
 * multi-normal distributed classes. This is an implementation of the Fisher
 * linear discrimination function, which has been extended with a quadratic
 * part. This function accepts on input a DATASET * pointer to a label set.
 * ClassA and ClassB specify for which class labels
 * the fisher function must be calculated. And c is defined as follows:
 *
 * C = P(ClassB)/P(ClassA);
 *
 * Where P(b) and P(a) are the apriori changes to select a member of class b and
 * class a respectively. The descision function is modeled as
 *        xT*W2*x + W1*x + b = 0,
 * where b is the constant term. On output W2[1..n][1..n], W1[1..n] and b
 * are set to the optimal value.
 *
 * Return value : TRUE  If an error is detected.
 *                FALSE If a valid fisher function is found.
 */

int fisherq_dataset (DATASET *lset, int ClassA, int ClassB, double c, double **W2, double *W1, double *b)
{
	double *M0, **C0, *M1, **C1;
	int nInputs;

	if (lset->SetFlag & IOSET)
	{
		sprerror("fisherq_dataset(): LABELSET expected.");
		return TRUE;
	}

	nInputs = (int) lset->NumInputs;

	if ((M0 = vector(1, nInputs)) == NULL)
		return TRUE;
	if ((C0 = matrix(1, nInputs, 1, nInputs)) == NULL)
	{
		free_vector(M0, 1, nInputs);
		return TRUE;
	}
	if (mean_var_dataset(lset, ClassA, M0, C0))
	{
		free_matrix(C0, 1, nInputs, 1, nInputs);
		free_vector(M0, 1, nInputs);
		return TRUE;
	}

	if ((M1 = vector(1, nInputs)) == NULL)
	{
		free_matrix(C0, 1, nInputs, 1, nInputs);
		free_vector(M0, 1, nInputs);
		return TRUE;
	}
	if ((C1 = matrix(1, nInputs, 1, nInputs)) == NULL)
	{
		free_matrix(C0, 1, nInputs, 1, nInputs);
		free_vector(M0, 1, nInputs);
		free_vector(M1, 1, nInputs);
		return TRUE;
	}
	if (mean_var_dataset(lset, ClassB, M1, C1))
	{
		free_matrix(C0, 1, nInputs, 1, nInputs);
		free_matrix(C1, 1, nInputs, 1, nInputs);
		free_vector(M0, 1, nInputs);
		free_vector(M1, 1, nInputs);
		return TRUE;
	}

	if (fisherq(C0, C1, M0, M1, c, nInputs, W2, W1, b))
	{
		free_matrix(C0, 1, nInputs, 1, nInputs);
		free_matrix(C1, 1, nInputs, 1, nInputs);
		free_vector(M0, 1, nInputs);
		free_vector(M1, 1, nInputs);
		return TRUE;
	}

	free_vector(M0, 1, nInputs);
	free_matrix(C0, 1, nInputs, 1, nInputs);
	free_vector(M1, 1, nInputs);
	free_matrix(C1, 1, nInputs, 1, nInputs);

	return FALSE;
}


#undef FQTINY
#undef QUADR

/*
 * End of Source.
 */
